"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import sys
import copy

from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.framework.configurable import Configurable, RawOSConfigurableCache
from panda3d_gemstone.framework.singleton import SingletonWrapper

from panda3d_gemstone.io.file_system import path_exists

from panda3d_gemstone.engine import runtime
from panda3d_gemstone.engine.prc import get_prc_value, load_prc_file_data
from panda3d_gemstone.engine.audio import SoundManager

from panda3d.core import ConfigVariableList

class InvalidEnvironmentConfig(RuntimeError):
    """
    Represents an invalid environment configuration
    """

class ApplicationOptions(Configurable, InternalObject):
    """
    Base class for all Gemstone framework option
    file management
    """

    THREADING_MODELS = ['App/Cull/Draw', 'Cull/Draw', '/Draw', 'Cull']
    PRC_VARIABLES = ['load-display', 'fullscreen', 'win-size', 'sync-video', 'show-frame-rate-meter']
    SUPPORTED_GRAPHICS_LIBRARIES = ConfigVariableList('gs-display')
    SUPPORTED_LANGUAGES = ConfigVariableList('gs-option-lang')

    def __init__(self, config_path, local_settings):
        Configurable.__init__(self, config_path, override_cache=RawOSConfigurableCache())
        InternalObject.__init__(self)
        self.local_settings = local_settings
        self.config_prc_settings = None
        self.__validate_environment()

    def __validate_environment(self) -> None:
        """
        Validates the application environment to ensure its 
        is properly configured
        """

        if len(ApplicationOptions.SUPPORTED_GRAPHICS_LIBRARIES) <= 0:
            raise InvalidEnvironmentConfig('Invalid environment; No graphics library options are configured')

        if len(ApplicationOptions.SUPPORTED_LANGUAGES) <= 0:
            raise InvalidEnvironmentConfig('Invalid environment; No application languages are configured')

    def set_options_before_panda_start(self) -> None:
        """
        Performs application settings changes prior to Showbase
        initialization
        """

        self.config_prc_settings = self.get_settings_from_prc(self.PRC_VARIABLES)
        self.set_video_options()
        self.set_interface_options()
        self.__set_platform_specific_options()

        if runtime.is_developer_build():
            application_name = runtime.application.get_application_name()
            self.set_engine_setting('pstats-name', '%s Performance Stats' % application_name)

    def __set_platform_specific_options(self) -> None:
        """
        Sets the platform specific options for the application
        """

        platform_name = sys.platform
        setter_name = 'set_%s_options' % platform_name
        if hasattr(self, setter_name):
            self.notify.info('Setting platform options for: %s' % platform_name)
            getattr(self, setter_name)()
        else:
            self.notify.warning('No platform options setter found for: %s' % platform_name)

    def set_darwin_options(self) -> None:
        """
        Sets the darwin operation system specific options
        """

        load_prc_file_data("gl-version 3 2")

    def set_win32_options(self) -> None:
        """
        Sets the windows operation system specific options
        """

        self.__set_threading_model_is_enabled()

    def set_linux2_options(self) -> None:
        """
        Sets the linux operation system specific options
        """

        self.__set_threading_model_is_enabled()

    def __set_threading_model_is_enabled(self) -> None:
        """
        Sets the threading model for the application
        """

        multithreaded = self.get_setting('enableMultithreading', False, self._valdiate_true_or_false)
        threading_model = self.get_setting('multithreadingModel', 0, self._validate_threading_model)
        if multithreaded:
            threading_value = self.THREADING_MODELS[threading_model]
            self.notify.debug('Using threading model: %s' % threading_value)
            self.set_engine_setting('threading-model', threading_value)

    def set_options_after_panda_start(self) -> None:
        """
        Sets the application's options post Panda3D setup
        """

        self.set_audio_options()

    def set_interface_options(self) -> None:
        """
        Sets the application's interface options
        """

        self.get_setting('activeLanguage', True, self._validate_language)

    def set_engine_setting(self, key: str, value: object) -> None:
        """
        Sets the Panda3D engine prc configuration value
        """

        if runtime.has_base():
            self.notify.warning('Attempting to set engine setting "%s" after ShowBase startup. Setting may not take effect' % (
                key))
        
        final = ''
        if isinstance(value, bool):
            final = '#t' if value else '#f'
        elif isinstance(value, tuple) or isinstance(value, list):
            parts = []
            for v in value:
                parts.append(str(v))
            final = ' '.join(parts)
        else:
            final = value
        
        self.notify.debug('Setting engine setting "%s" to "%s"' % (key, final))
        load_prc_file_data('%s %s' % (key, final), 'application-setting')

    def set_video_options(self) -> None:
        """
        Sets the application's video and graphics options
        """

        display_default = None
        fullscreen_default = None
        size_default = None
        sync_video_default = None
        frame_rate_default = False

        if self.config_prc_settings:
            display_default = self.config_prc_settings.get('load-display', '')
            display_default = display_default[5:] if display_default.startswith('panda') else None
            
            sync_video_default = self.config_prc_settings.get('sync-video', 1)
            sync_video_default = sync_video_default == True or sync_video_default == '#t' or sync_video_default == 't'

            frame_rate_default = self.config_prc_settings.get('show-frame-rate-meter', False)
            frame_rate_default = frame_rate_default == True or frame_rate_default == '#t' or frame_rate_default == 't'
            
            #TODO: load the rest of the defaults

        display = self.get_setting('display', display_default, self._validate_display)
        fullscreen = self.get_setting('fullscreen', fullscreen_default, self._valdiate_true_or_false)
        size = self.get_setting('displaySize', size_default)
        sync_video = self.get_setting('syncVideo', sync_video_default, self._valdiate_true_or_false)
        frame_rate = self.get_setting('enableFrameRate', frame_rate_default, self._valdiate_true_or_false)

        self.set_engine_setting('fullscreen', fullscreen)
        self.set_engine_setting('load-display', 'panda%s' % display)
        self.set_engine_setting('win-size', size)
        self.set_engine_setting('sync-video', sync_video)
        self.set_engine_setting('show-frame-rate-meter', frame_rate)
 
    def _validate_value(self, val: object, options: list) -> bool:
        """
        Validates the value and returns true if the value is 
        contained in the options list
        """

        return val in options

    def _validate_type(self, val: str, val_type: object) -> bool:
        """
        Validates the value's type and returns true if it matches the
        type provided in val_type
        """

        return type(val) == val_type

    def _validate_threading_model(self, val: int) -> bool:
        """
        Validates the value to ensure its a valid threading model
        option
        """

        return val >= 0 and val < len(self.THREADING_MODELS)

    def _validate_display(self, val: str) -> bool:
        """
        Validates the value to ensure its a valid graphics library
        option
        """

        return self._validate_value(val, self.SUPPORTED_GRAPHICS_LIBRARIES)

    def _validate_language(self, val: str) -> bool:
        """
        Validates the value to ensure its a valid language option
        """

        return self._validate_value(val, self.SUPPORTED_LANGUAGES)

    def _validate_between_zero_and_one(self, val: int) -> bool:
        """
        Returns true if the valie is between 0 and 1
        """

        return val >= 0 and val <= 1

    def _valdiate_true_or_false(self, val: bool) -> bool:
        """
        Returns true if the value is a valid boolean
        """

        return type(val) is bool

    def _validate_string(self, val: str) -> bool:
        """
        Returns true if the value is a valid string type
        """

        return self._validate_type(val, str)

    def set_audio_options(self) -> None:
        """
        Sets the application's audio options
        """

        enable_music = self.get_setting('enableMusic', True, self._valdiate_true_or_false)
        volume_music = 0
        if enable_music:
            volume_music = self.get_setting('volumeMusic', 0.1, self._validate_between_zero_and_one)
        SoundManager.get_singleton().set_music_volume(volume_music)

        enable_sfx = self.get_setting('enableSFX', True, self._valdiate_true_or_false)
        volume_sfx = 0
        if enable_sfx:
            volume_sfx = self.get_setting('volumeSFX', 1.0, self._validate_between_zero_and_one)
        SoundManager.get_singleton().set_sound_effects_volume(volume_sfx)

    def get_setting(self, setting_name: str, default_fallback: object = None, valdiation_callback: object = None) -> object:
        """
        Retrieves an application setting value and runs it against a validation
        callback if present.
        """

        if setting_name in self.local_settings:
            val = self.local_settings.get(setting_name)
            is_ok = True

            if valdiation_callback is not None:
                is_ok = valdiation_callback(val)

            if is_ok:
                return val
            else:
                self.notify.warning('Invalid value %s for option %s found in application settings' % (
                    val, setting_name))

        if setting_name in self.configuration:
            ret = self.configuration.get(setting_name)
            self.local_settings[setting_name] = ret

            return ret

        if default_fallback is not None:
            self.notify.info('Setting default setting for "%s": %s' % (setting_name, default_fallback))
            self.local_settings[setting_name] = default_fallback

        self.notify.warning('No default setting for "%s" specified in %s. Defaulting to %s' % (
            setting_name, self.path, default_fallback))
        
        return default_fallback

    def get_settings_from_prc(self, setting_name_list) -> dict:
        """
        Retrieves the required settings from the Panda3D runtime config
        """

        values = {}
        for setting_name in setting_name_list:
            result = get_prc_value(setting_name, None)
            if result:
                values[setting_name] = result
            else:
                self.notify.warning('Failed to retrieve prc value for "%s"' % setting_name)

        return values

class LocalSettings(dict, Configurable):
    """
    Represents a local application configuration file
    """

    def __init__(self, config_path: str):
        dict.__init__(self)
        self.__child_dicts = {}
        self.__first_time = True
        self.__dirty = False

        if os.path.exists(config_path):
            self.__first_time = False

        Configurable.__init__(self, config_path, override_cache=RawOSConfigurableCache())
        self.update(self.configuration)
        del self.configuration
    
    def is_first_time(self) -> bool:
        """
        Returns true if the local settings is a newly created
        file
        """

        return self.__first_time

    def is_dirty(self) -> bool:
        """
        Returns true if the application settings are currently dirty
        """

        return self.__dirty

    def write(self) -> None:
        """
        Writes the application settings to file
        """

        f = open(self.path, 'w')
        f.write('[Configuration]\n')

        for key, value in list(self.items()):
            f.write('%s: %s\n' % (str(key), str(value)))

        for key, value in self.__child_dicts.items():
            f.write('\n[%s]\n' % str(key))
            if type(value) == dict:
                for subKey, sub_value in value.items():
                    if type(sub_value) == type(''):
                        f.write("%s: '%s'\n" % (str(sub_key), str(sub_value.encode('utf8'))))
                    elif type(subValue) == str:
                        f.write("%s: '%s'\n" % (str(sub_key), str(sub_value)))
                    else:
                        f.write('%s: %s\n' % (str(sub_key), str(sub_value)))

        f.close()     
        self.__dirty = False     

    def load_data(self, section: str, data: dict) -> None:
        """
        Loads the data from the local application settings file
        """

        if data is None:
            data = {}

        self.__child_dicts[section] = data

    def reset(self) -> None:
        """
        Resets the local settings
        """

        if os.path.exists(self.path):
            os.remove(self.path)

        if os.path.exists(self.path):
            os.remove(self.path)

        self.__first_time = True
        self.clear()

    def set_value_in_child_dict(self, child_dict_name: object, key: str, value: object) -> None:
        """
        """

        temp_dict = self.__child_dicts.get(child_dict_name, {})
        temp_dict[key] = value
        self.__child_dicts[child_dict_name] = temp_dict

    def get_value_in_child_dict(self, child_dict_name: object, key: str, value: object) -> None:
        """
        """

        return self.__child_dicts.get(child_dict_name, {}).get(key, default)

    def get_child_dicts_copy(self) -> object:
        """
        """

        return copy.deepcopy(self.__child_dicts)

    def replace_child_dict(self, new_dict: dict) -> None:
        """
        """

        self.__child_dicts = dict(new_dict)

    def on_setting_changed(self, key: str, new_value: object, old_value: object) -> None:
        """
        Called on setting changed
        """

        self.__dirty = True

    def __setitem__(self, key: str, item: object) -> None:
        """
        Custom item setter for calling on_setting_changed
        """

        old_value = dict.get(self, key, None)
        dict.__setitem__(self, key, item)
        self.on_setting_changed(key, item, old_value)

class SingletonLocalSettings(SingletonWrapper):
    """
    Singleton wrapper for the application local settings
    """

    WRAPPED_OBJECT = LocalSettings


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

try:
    from setuptools import setup as __setup
except ImportError:
    from distutils.core import setup as __setup

import os
import re
import sys
import codecs
import subprocess

# Verify the python version meets minimum requirements
if sys.version_info[0] == 3 and sys.version_info < (3, 4):
    sys.exit("Error: Python 3.4 or newer is required.")

__here = os.path.abspath(os.path.dirname(__file__))
__root_package = 'panda3d_gemstone'
__version_file = '__version__.py'
__description = 'top level framework built for the Panda3D game engine'
__license = 'MIT'
__repository = 'https://github.com/thetestgame/panda3d-gemstone'
__classifiers = [
    'Development Status :: 2 - Pre-Alpha',
    'Natural Language :: English',
    'Programming Language :: Python :: 3'
]

def __has_thirdparty(import_string: str) -> bool:
    """
    Returns true if the import string is valid
    """

    found = False
    import importlib
    try:
        module = importlib.import_module(import_string)
        found = True
    except:
        pass

    return found

def __codec_read(*parts) -> object:
    """
    Reads a file with the codedcs module
    """

    with codecs.open(os.path.join(__here, *parts), 'r') as fp:
        return fp.read()

def __get_file_contents(file_path: str) -> str:
    """
    Returns the contents of a file as a string
    """

    return open(file_path).read()

def __get_readme() -> str:
    """
    Returns the contents of the package's README file
    """

    return __get_file_contents('README.md')

def __get_requirements(file_path: str = 'requirements.txt') -> list:
    """
    Returns the contents of the requested requirements.txt
    file
    """

    try: # for pip >= 10
        from pip._internal.req import parse_requirements
    except ImportError: # for pip <= 9.0.3
        from pip.req import parse_requirements

    install_reqs = parse_requirements(file_path, session='hack')
    reqs = [str(ir.requirement) for ir in install_reqs]
    
    return reqs

def __find_version(*file_paths) -> str:
    """
    Attempts to find the version from a specific module path
    """

    version_file = __codec_read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)

    if version_match:
        return version_match.group(1)

    raise RuntimeError('Install failed. Unable to find version string')

def __list_as_string(parts: list, splice: str = ', ') -> str:
    """
    Returns the list as a string
    """

    return splice.join(parts)

def __get_authors() -> str:
    """
    Returns all authors as a string
    """

    authors = ['Jordan Maxwell']
    #results = subprocess.check_output(['git', 'log', '--pretty="%an"']).decode()
    #results = results.split('\n')

    #for author in results:
    #    author = author.replace('"', '')
    #    if author not in authors and author != '':
    #        authors.append(author)

    return __list_as_string(authors)

def __get_modules(root_module: str) -> list:
    """
    Retrieves all modules inside a root module
    """

    modules = [root_module]
    for root, subdirs, _ in os.walk(root_module):
        root = root.replace(os.sep, '.')
        for subdir in subdirs:
            if subdir.startswith('__'):
                continue

            modules.append('%s.%s' % (root, subdir))

    return modules

def __main() -> int:
    """
    Main entry point for the setup script
    """

    __setup(
        name=__root_package,
        description=__description,
        long_description=__get_readme(),
        license=__license,
        version=__find_version(__root_package, __version_file),
        author=__get_authors(),
        url=__repository,
        #install_requires=__get_requirements(),
        packages=__get_modules(__root_package),
        zip_safe=False,
        classifiers=__classifiers)

    return 0

if __name__ == '__main__':
    sys.exit(__main())


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import builtins

from direct.showbase.ShowBase import ShowBase

from panda3d_gemstone.framework.service import Service, Command
from panda3d_gemstone.logging import utilities as logging
from panda3d_gemstone.engine import runtime, prc

from panda3d.core import WindowProperties, AntialiasAttrib

class GemstoneShowBase(Service, ShowBase):
    """
    Custom Gemstone framework varient of the stock
    Panda3D ShowBase object
    """ 

    def __init__(self, *args, **kwargs):
        Service.__init__(self, initialize_handle_attributes_as_commands=False)
        ShowBase.__init__(self, *args, **kwargs)

        self.config = prc.ShowbaseConfig()
        builtins.config = self.config
        runtime.config = self.config

    def setup(self) -> None:
        """
        Performs setup operations on the showbase instance
        """

    def destroy(self) -> None:
        """
        Performs destruction operations on the showbase instance
        """

        Service.destroy(self)
        ShowBase.destroy(self)

    def set_exit_callback(self, func: object) -> None:
        """
        Sets the showbase's shutdown callback
        """

        assert func != None
        assert callable(func)

        self.exitFunc = func

class ApplicationShowBase(GemstoneShowBase):
    """
    Standard ShowBase instance for the Gemstone framework.
    Used by most application use cases
    """

    notify = logging.get_notify_category('showbase')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setup()

        self.set_antialias(prc.get_prc_bool('gs-panda-antialias', True))
        self.__showing_frame_rate = prc.get_prc_bool('show-frame-rate-meter', False)

        self.add_command(Command('toggle frame rate', self.toggle_frame_rate))

    def setup(self) -> None:
        """
        Performs setup operations on the showbase instance
        """

        super().setup()

        self.__setup_runtime()

    def destroy(self) -> None:
        """
        Performs destruction operations on the showbase instance
        """

        super().destroy()

    def __setup_runtime(self) -> None:
        """
        Performs runtime variable setup operations
        """

        runtime.base = self
        runtime.task_mgr = self.task_mgr
        runtime.loader = self.loader
        runtime.cam = self.cam
        runtime.camera = self.camera
    
    def set_window_title(self, window_title: str) -> None:
        """
        Sets the primary window's title
        """

        # Verify we have a window instance
        if not self.win:
            return

        props = WindowProperties()
        props.set_title(window_title)
        self.win.request_properties(props)

    def set_window_dimensions(self, origin: tuple, size: tuple) -> None:
        """
        Sets the current window dimensions
        """

        if not self.win:
            return

        props = WindowProperties()
        props.set_origin(*origin)
        props.set_size(*size)
        self.win.request_properties(props)

    def get_window_dimensions(self) -> tuple:
        """
        Returns the current windows dimensions
        """

        origin = (-1, -1)
        size = (-1, -1)

        if not self.win:
            return (origin, size)

        props = self.win.get_properties()
        if sys.platform == 'darwin':
            origin = (25, 50)
        elif props.has_origin():
            origin = (props.get_x_origin(), props.get_y_origin())
        
        if props.has_size():
            size = (props.get_x_size(), props.get_y_size())

        return (origin, size)

    def set_clear_color(self, clear_color: object) -> None:
        """
        Sets the primary window's clear color
        """

        # Verify we have a window instance
        if not self.win:
            return

        self.win.set_clear_color(clear_color)

    def set_antialias(self, antialias: bool) -> None:
        """
        Sets the graphics library based antialiasing state
        """

        if not prc.get_prc_bool('framebuffer-mutlisample', False):
            prc.set_prc_value('framebuffer-multisample', True)

        if prc.get_prc_int('multisamples', 0) < 2:
            self.notify.warning('Multisamples not set. Defaulting to a value of 2')
            prc.set_prc_value('multisamples', 2)

        if antialias:
            self.render.set_antialias(AntialiasAttrib.MAuto)
        else:
            self.render.clear_antialias()

    def post_window_setup(self) -> None:
        """
        Performs setup operations after the window has succesfully
        opened
        """

    def open_default_window(self) -> object:
        """
        Opens a window with the default configuration
        options
        """

        props = WindowProperties.get_default()
        return self.openMainWindow(props = props)

    def openMainWindow(self, *args, **kwargs) -> object:
        """
        Custom override of the ShowBase openMainWindow function
        for handling runtime registering
        """

        result = ShowBase.openMainWindow(self, *args, **kwargs)

        if result:
            runtime.window = self.win
            runtime.render = self.render

            self.post_window_setup()

        return self.win

    def toggle_frame_rate(self) -> None:
        """
        Toggles the application's frame rate meter
        """

        self.__showing_frame_rate = not self.__showing_frame_rate
        self.set_frame_rate_meter(self.__showing_frame_rate)

    def is_oobe(self) -> bool:
        """
        Returns true if the ShowBase instance is in oobe mode
        """

        if not hasattr(self, 'oobeMode'):
            return False

        return self.oobeMode

class HeadlessApplicationShowbase(ApplicationShowBase):
    """
    Headless varient of the ApplicationShowBase object. Creates
    without a primary window instance
    """

    def __init__(self, *args, **kwargs):
        kwargs['window_type'] = 'none'
        super().__init__(*args, **kwargs)


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import sys
import logging

from direct.gui.DirectGui import *

from panda3d.core import Filename, WindowProperties, CullBinManager

from panda3d_gemstone.framework.configurable import Configurable
from panda3d_gemstone.framework.singleton import Singleton

from panda3d_gemstone.engine.model import TextureModel
from panda3d_gemstone.engine import runtime

class MouseCursor(Configurable, Singleton):
    """
    """

    class Cursor(object):
        """
        Represents a cursor loaded into memory
        """

        def __init__(self, filename: object, frame: object):
            self.filename = filename
            self.frame = frame

    def __init__(self, config_path: str, window: object, window_properties_class: object = WindowProperties):
        Singleton.__init__(self)
        self.__hardware_cursor = True
        self.__software_cursor = False
        self.__cursors = {}
        self.__window_properties = window_properties_class()
        self.__window = window
        self.__current = None
        self.__current_name = ''
        Configurable.__init__(self, config_path)
        self.initialize()

    @property
    def current_name(self) -> str:
        """
        Returns the mouse cursor's current 
        cursor name as a property
        """

        return self.get_current()

    @current_name.setter
    def current_name(self, cursor_name: str) -> None:
        """
        Sets the mouse cursor's current
        cursor name as a setter
        """

        self.set_current(cursor_name)

    def get_current(self) -> str:
        """
        Returns the current cursor name
        """

        return self.__current_name

    def set_current(self, cursor_name: str, reset: bool = False) -> None:
        """
        Sets the current cursor name
        """

        if self.__current_name == cursor_name and not reset:
            return

        if not cursor_name:
            if self.__hardware_cursor and self.__window:
                self.__window_properties.clear_cursor_filename()
                self.__window.request_properties(self.__window_properties)
            
            if self.__software_cursor:
                if runtime.base.mouseWatcherNode:
                    runtime.base.mouseWatcherNode.clear_geometry()
                if self.__current:
                    self.__current.stash()
        elif self.has_cursor(cursor_name):
            self.__current_name = cursor_name
            if self.__hardware_cursor and self.__window:
                self.__window_properties.set_cursor_filename(Filename(self.get_cursor_filename(cursor_name)))
                self.__window.request_properties(self.__window_properties)

            if self.__software_cursor:
                if self.__current:
                    self.__current.stash()

                self.__current = self.get_cursor_frame(cursor_name)
                runtime.base.mouseWatcherNode.set_geometry(self.__current.node())
                self.__current.unstash()
        else:
            self.notify.warning('Failed to set cursor. Unknown cursor "%s"' % cursor_name)

    def add_cursor(self, cursor_name: str, filename: str) -> None:
        """
        """

        cursor = TextureModel(filename + '.ini')
        if cursor.get_num_children() and cursor._models:
            cursor_frame = DirectFrame(geom=cursor.get_child(0), relief=None, parent=runtime.base.render2d, sort_order=10000)
            cursor_frame.stash()

            self.__cursor[cursor_name] = MouseCursor.Cursor(cursor._models[0] + '.ico')
            if self.__cursors[cursor_name] is None:
                self.notify.error('Failed to load cursor: %s' % cursor_name)
        else:
            self.notify.error('Failed to load cursor: %s; Invalid geometry' % cursor_name)

    def has_cursor(self, cursor_name: str) -> bool:
        """
        """

        return cursor_name in self.__cursors

    def get_cursor_filename(self, cursor_name: str) -> str:
        """
        """

        if self.has_cursor(cursor_name):
            return self.__cursors[cursor_name].filename

        return ''

    def get_cursor_frame(self, cursor_name: str) -> object:
        """
        """

        if self.has_cursor(cursor_name):
            return self.__cursors[cursor_Name].filename

        return None

    def load_cursors_data(self, data: dict) -> None:
        """
        """

        keys = list(data.keys())
        keys.sort()

        for key in keys:
            self.add_cursor(key, data[key])

    def set_use_hardware_cursor(self, state: bool) -> None:
        """
        """

        self.__hardware_cursor = state

    def set_use_software_cursor(self, state: bool) -> None:
        """
        """

        self.__software_cursor = state


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from panda3d_gemstone.engine import runtime

from panda3d_gemstone.framework.singleton import Singleton
from panda3d_gemstone.engine.prc import get_prc_bool, get_prc_string
from panda3d_gemstone.engine.prc import get_prc_int
from panda3d_gemstone.engine import runtime
from panda3d_gemstone import __version__ as gemstone_version

class ApplicationVersion(Singleton):
    """
    Wrapper to identify the version of the Gemstone Application
    """

    def __init__(self):
        version_parts = gemstone_version.split('.')
        self._major = get_prc_int('gs-v-major', int(version_parts[0]))
        self._minor = get_prc_int('gs-v-minor', int(version_parts[1]))
        self._release = get_prc_int('gs-v-release', int(version_parts[2]))
        super().__init__()

        use_sc_revision = get_prc_bool('gs-sc-revision', False)
        if use_sc_revision and runtime.is_developer_build():
            self._release = self.__get_version_from_revision()

    def major(self) -> int:
        """
        Application's major version integer
        """

        return self.major

    @property
    def minor(self) -> int:
        """
        Applications's minor version integer
        """

        return self._minor

    @property
    def release(self) -> int:
        """
        Application's release version integer
        """
        
        return self._release

    def __str__(self) -> str:
        """
        Returns the version as a string
        """

        base_str = '%s.%s.%s' % self.get()

        is_dev = runtime.is_developer_build()
        if is_dev:
            base_str = '%s - DEV' % base_str
    
        return base_str

    def get(self) -> tuple:
        """
        Returns the version as a tuple
        """

        return (self._major, self._minor, self._release)

    def __get_version_from_revision(self) -> str:
        """
        Returns the applications version from its current source
        control revision
        """

        from panda3d_gemstone.tools import source_control
        return source_control.get_local_head_revision('.')


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import sys

from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from PyQt5.QtGui import *

from panda3d_gemstone.engine import prc

from panda3d_gemstone.editor.widgets import panda, statusbar, menu, menubar
from panda3d_gemstone.framework.internal_object import InternalObject

class EditorWindow(QMainWindow, InternalObject):
    """
    Primary window/widget for the PyQt5 editor application
    """

    def __init__(self, application: object, *args, **kwargs):
        super(QMainWindow, self).__init__(*args, **kwargs)
        super(InternalObject, self).__init__()

        self._application = application
        self._panels = {}

        self.build_window()

    def build_window(self) -> None:
        """
        Constructs the primary editor window
        """

        # Set window info
        window_origin = prc.get_prc_string('window-origin', '0,0').split(',')
        window_size = prc.get_prc_string('window-size', '800,600').split(',')
        self.setGeometry(
            int(window_origin[0]), 
            int(window_origin[1]), 
            int(window_size[0]), 
            int(window_size[1]))
        self.setWindowIcon(QIcon('data/EditorData/images/logo.png'))

        # Load widgets
        self.menu_bar = menubar.WindowMenuBar(self)
        self.status_bar = statusbar.WindowStatusBar(self)

        self.load_menu_bar()
        self.load_panda_widget()

        # Finish construction
        self.finalize_window()

    def finalize_window(self) -> None:
        """
        Finalizes the newly created window
        """

        self.panda.force_resize()

    def load_panda_widget(self) -> None:
        """
        Loads the Panda3D based widget
        """

        self.panda = panda.QTPandaWidget()
        self.panda.setup()
        self.setCentralWidget(self.panda)

    def load_menu_bar(self) -> None:
        """
        Loads the primary window's menu bar
        """

        self.file_menu = menu.EditorFileMenu(self, self.menu_bar)
        self.edit_menu = menu.EditorEditMenu(self, self.menu_bar)
        self.view_menu = menu.EditorViewMenu(self, self.menu_bar)
        self.debug_menu = menu.EditorDebugMenu(self, self.menu_bar)
        self.help_menu = menu.EditorHelpMenu(self, self.menu_bar)

    def toggle_panel(self, panel_class: object, state: bool, dock: object = Qt.RightDockWidgetArea) -> None:
        """
        Toggles the requested panel class
        """

        panel_key = panel_class.__name__
        existing = self._panels.get(panel_key, None)
        self.notify.debug('Toggling Panel: %s' % panel_key)
        if existing and not state:
            self.removeDockWidget(existing)
            del self._panels[panel_key]
        elif not existing and state:
            widget = panel_class(self)
            self.addDockWidget(dock, widget)
            self._panels[panel_key] = widget

    def get_panel(self, panel_class: str) -> object:
        """
        Retrieves the current panel instance if one exists
        """

        panel_key = panel_class.__name__
        return self._panels.get(panel_key, None)


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.engine import shader, texture, runtime
from panda3d_gemstone.world.entity import Entity

from panda3d.core import ShaderTerrainMesh, GeoMipTerrain, Filename
from panda3d.core import Vec3

class TerrainBase(Entity):
    """
    Base class of all Panda3D terrain objects
    """

    def __init__(self, config_path: str, terrain: object = None):
        self.__texture = None
        self.__heightfield_texture = None
        self.__terrain_scale = Vec3(1, 1, 1)
        assert terrain != None
        self.__terrain = terrain
        self.__values = {}
        super().__init__(config_path)
        self.configure_terrain(self.__values)

    @property
    def terrain(self) -> object:
        """
        Panda3D native object instance
        """

        return self.__terrain

    def configure_terrain(self, values) -> None:
        """"
        Performs terrain configuration operations
        """

        # Configure terrain
        for value_key in self.__values:
            if hasattr(self.__terrain, value_key):
                value = self.__values[value_key]
                func = getattr(self.__terrain, value_key)
                if func and False:
                    if isinstance(value, tuple):
                        func(*value)
                    else:
                        func(value)
                else:
                    attr = getattr(self.__terrain, value_key.replace('set_', ''))
                    if attr:
                        attr = value
                    else:
                        self.notify.warning('%s does not have attribute %s!' % (
                            self.__terrain.__class__.__name__, value_key))   

    def set_heightfield(self, heightfield_path: str) -> None:
        """
        Sets the terrain's heightfield texture
        """

        self.__heightfield_texture = texture.load_heightfield(heightfield_path)
        assert self.__heightfield_texture != None

    def get_heightfield(self) -> object:
        """
        Returns the terrain's heightfield texture
        """

        return self.__heightfield_texture

    def set_terrain_texture(self, texture_path: str) -> None:
        """
        Sets the terrain's base texture 
        """

        self.__texture = texture.load_texture(texture_path, okMissing=True)

    def get_terrain_texture(self) -> object:
        """
        Retrieves the terrain's base texture
        """

        return self.__texture
 
    def set_terrain_scale(self, scale: object) -> None:
        """
        Sets the terrain's scale value
        """

        inputs = []
        if isinstance(scale, tuple):
            inputs = scale
        else:
            for num in scale.split(','):
                inputs.append(float(num))

        self.__terrain_scale = Vec3(*inputs)

    def get_terrain_scale(self) -> None:
        """
        Returns the terrain's scale value
        """

        return self.__terrain_scale

    def __get_value(self, key: str) -> object:
        """
        Attempts to retrieve the value from the values dictionary
        """

        return self.__values.get(key, None)

    def __set_value(self, key: str, obj: object) -> None:
        """
        Sets the value in the values dictionary
        """

        self.__values[key] = obj

    def __getattr__(self, key: str) -> object:
        """
        Custom attribute getter for passing along function
        calls to the native Panda3D terrain object
        """

        result = None

        def has_attribute(obj: object, key: str) -> bool:
            """
            Returns true if the object has a attribute from the getter/setter key
            """

            attrib_key = key.replace('get_', '').replace('set_', '')
            return hasattr(obj, key) or hasattr(obj, attrib_key)

        def valid_get(obj: object, key: str) -> bool:
            """
            Returns true if the requested attribute is a valid getter
            """

            ignored = ['get_class_type']
            return key.startswith('get_') and key not in ignored
        
        if isinstance(key, str):
            if has_attribute(self.__terrain, key):
                if valid_get(self.__terrain, key):
                    result = lambda: self.__get_value(key)
                elif key.startswith('set_'):
                    result = lambda x: self.__set_value(key, x)
                else:
                    return getattr(self.__terrain, key)

        if result is None:
            raise AttributeError('%s instance does not have attribute %s' % (
                self.__class__.__name__, key))

        return result

class CPUTerrain(TerrainBase):
    """
    Represents a gemstone cpu terrain object
    by wrapping the Panda3D native GeoMipTerrain
    node object
    """

    def __init__(self, config_path: str):
        self.__terrain_texture = None
        super().__init__(config_path, GeoMipTerrain(self.__class__.__name__))

        # Finalize terrain node
        self.terrain.set_heightfield(self.get_heightfield())
        self.__terrain_node = self.__terrain.get_root()
        self.__terrain_node.reparent_to(self)
        height_scale = self.get_scale().get_z()
        self.__terrain_node.set_scale(1, 1, height_scale)
        self.__terrain_node.set_z(self.__terrain_node.get_z() - (height_scale / 2))
        if self.get_terrain_texture():
            self.__terrain_node.set_texture(self.get_terrain_texture())
        self.__terrain.generate()

    @property
    def terrain_node(self) -> object:
        """
        Panda3D native Node instance representing
        the GeoMipTerrain instance
        """

        return self._terrain_node

class GPUTerrain(TerrainBase):
    """
    Represents a gemstone gpu terrain object
    by wrapping the Panda3d native ShaderTerrainMesh
    node object
    """

    def __init__(self, config_path: str):
        self.__terrain_texture = None
        self.__target_triangle_width = 10
        self.__vertex_shader = ''
        self.__fragment_shader = ''

        super().__init__(config_path, ShaderTerrainMesh())

        # Finalize terrain
        self.terrain.heightfield = self.get_heightfield()
        self.terrain.generate()

        self.__terrain_node = self.attach_new_node(self.terrain)
        terrain_scale = self.get_terrain_scale()
        self.__terrain_node.set_scale(terrain_scale)
        self.__terrain_node.set_pos(0, 0, -terrain_scale.get_z()/2)
        if self.get_terrain_texture():
            self.__terrain_node.set_texture(self.get_terrain_texture(), 1)

        self.__terrain_shader = shader.load_shader(self.__vertex_shader, self.__fragment_shader)
        self.__terrain_node.set_shader(self.__terrain_shader)
        self.__terrain_node.set_shader_input('camera', runtime.camera)
    
    @property
    def terrain_node(self) -> object:
        """
        Panda3D native NodePath representing the ShaderTerrainMesh
        instance
        """

        return self.__terrain_node

    def set_terrain_fragment_shader(self, shader_path: str) -> None:
        """
        Sets the terrain's fragment shader path
        """

        self.__fragment_shader = shader_path

    def get_terrain_fragment_shader(self) -> str:
        """
        Returns the terrain's fragment shader path
        """

        return self.__fragment_shader

    def set_terrain_vertex_shader(self, shader_path: str) -> None:
        """
        Sets the terrain's vertex shader path
        """

        self.__vertex_shader = shader_path

    def get_terrain_vertex_shader(self) -> str:
        """
        Returns the terrain's vertex shader path
        """

        return self.__vertex_shader

class PaintedChannel(InternalObject):
    """
    """

    def __init__(self):
        super().__init__()

class PaintedChannelTerrain(GPUTerrain):
    """
    """

    def __init__(self, config_path: str):
        super().__init__(config_path)
        self.__channels = []


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from panda3d_gemstone.framework.exceptions import verify_thirdparty_modules
verify_thirdparty_modules(['watchdog', 'limeade'])

from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.framework.utilities import disallow_production
from panda3d_gemstone.framework.singleton import Singleton
from panda3d_gemstone.engine import prc, runtime

from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

import traceback
import limeade

class ReloaderFileEventHandler(FileSystemEventHandler):
    """
    Custom FileSystemEventHandler instance for performing os file watch tasks
    for the ApplicationReloader instance
    """

    def __init__(self, reloader: object, callback: object):
        assert callable(callback)
        self.__reloader = reloader
        self.__callback = callback
        super().__init__()

    def __is_valid_file(self, file_path: str) -> bool:
        """
        Returns true if the file path relates to a file type we are listening
        for. Otherwise returns False
        """

        valid = False
        for file_type in list(self.__reloader.watched_types):
            if file_path.endswith('.%s' % file_type):
                valid = True
                break

        return valid

    def on_any_event(self, event: object) -> None:
        """
        Handles all incoming events from the watchdog module and checks
        if the event relates to a file we care about before firing the 
        requested callback provided by the ApplicationReloader instance
        """

        changed_path = event.src_path
        if self.__is_valid_file(changed_path):
            ext = changed_path.split('.')[-1]
            self.__callback(changed_path, ext) 

class ApplicationReloader(Singleton, InternalObject):
    """
    Development utility object for hot reloading a Gemstone application
    when a file change is detected.
    """

    def __init__(self):
        Singleton.__init__(self)
        InternalObject.__init__(self)
        self.notify.setInfo(True)
        self.__watched_types = prc.get_prc_list('gs-reload-type')
        self.__watch_path = prc.get_prc_string('gs-reload-watch-path', '.')
        self.__reloaded = False
        self.__observer = None
        self.__setup_watchdog()
    
    @property
    def watched_types(self) -> list:
        """
        ApplicationReloader's currently watched file extensions
        """

        return self.__watched_types

    def add_watch_type(self, ext: str) -> bool:
        """
        Attempts to add a new file watch type to the reloader.
        Returning True if the type was added succesfully. Otherwise
        returning False
        """

        if ext in self.__watched_types:
            self.notify.warning('Attempted to add already registered watch type: %s' % ext)
            return False

        self.__watched_types.append(ext)
        return True

    def remove_watch_type(self, ext: str) -> None:
        """
        Attempts to remove the watch type from the reloader.
        """

        if not ext in self.__watched_types:
            return

        self.__watched_types.remove(ext)

    def destroy(self) -> None:
        """
        Destroys the ApplicationReloader instance
        """

        self.deactivate()
        Singleton.destroy(self)
        PandaBaseObject.destroy(self)

    def activate(self) -> None:
        """
        Activates the ApplicationReloader instance
        """

        # Verify we are valid
        if not self.is_valid():
            self.notify.warning('Failed to activate %s. Invalid configuration detected.' % (
                self.__class__.__name__))

        if self.__observer:
            self.__observer.start()

    def deactivate(self) -> None:
        """
        Deactivates the ApplicationReloader instance
        """

        if self.__observer:
            self.__observer.stop()

    def has_reloaded(self) -> bool:
        """
        Returns true if the application has reloaded to log any potential
        instabilities caused by the reload.
        """

        return self.__reloaded

    def is_valid(self) -> bool:
        """
        Returns true if the application reloader has a valid
        configuration
        """

        return len(self.__watched_types) > 0
    
    @disallow_production
    def send_event(self, file_path: str, ext: str) -> None:
        """
        Broadcasts a Panda Messenger event to notify other components
        that a file has changed
        """

        runtime.messenger.send('gs-file-changed', [file_path, ext])

    @disallow_production
    def on_reload(self, file_path: str, ext: str) -> None:
        """
        Default on reload handler
        """

        self.notify.info('Detecting file change: %s' % file_path)

    @disallow_production
    def on_py_reload(self, file_path: str) -> None:
        """
        Handles Python file reload events
        """

        self.notify.info('Detecting source change in file: %s. Reloading' % (
            file_path))

        # Attempt to reload source. Catching any potential errors
        # that occure during reload
        try:
            limeade.refresh()
        except Exception as e:
            print('== An Error Occured Reloading Source ==')
            print(traceback.format_exc())

    @disallow_production
    def __on_file_changed(self, file_path: str, ext: str) -> None:
        """
        Called when a file change is detected
        """

        # Attempt to call handler if present
        handler_name = 'on_%s_reload' % ext
        has_handler = hasattr(self, handler_name)
        if has_handler:
            getattr(self, handler_name)(file_path)
        else:
            self.on_reload(file_path, ext)

        # Flag that we have reloaded
        if not self.__reloaded:
            self.__reloaded = True

        # Send application wide event
        self.send_event(file_path, ext)

    @disallow_production
    def __setup_watchdog(self) -> None:
        """
        Performs Watchdog setup operations
        """

        event_handler = ReloaderFileEventHandler(
            reloader=self,
            callback=self.__on_file_changed)
        self.__observer = Observer()
        self.__observer.schedule(event_handler, self.__watch_path, recursive=True)



"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from direct.gui.OnscreenText import OnscreenText

from panda3d.core import TextNode, Vec2, Vec3, Vec4
from panda3d.core import Point2, Point3, load_prc_file_data

from panda3d_gemstone.framework.runnable import Runnable
from panda3d_gemstone.io.file_system import path_exists

from panda3d_gemstone.engine.text import Font

def create_node_debug_hud(parent: object, pos: object, scale: object, node: object, font: str = 'data/fonts/framd.ini', labels: dict = {}) -> Inspector:
    """
    Creates a new on screen debug hud for a Panda3D NodePath
    """

    inspector = Inspector(fontPath=font, updateTime=0.25, parent=parent, pos=pos, scale=scale)

    for label_key in labels:
        inspector.add_object(label_key, labels[label_key], show_label=True)

    inspector.add_object('Pos', node.get_pos)
    inspector.add_object('heading', node.get_h)
    inspector.activate()

    return inspector

class InspectObject(object):
    """
    """

    def __init__(self, always_show: bool, show_label: bool, callback: object, *args):
        self.__always_show = always_show
        self.__show_label = show_label

        if callable(callback):
            self.__callback = callback
        else:
            self.__callback = lambda callback = callback, *args: callback
        self.__args = args

    @property
    def value(self) -> tuple:
        """
        Returns the inspect object's value as a property
        """

        return self.get_value()

    def get_value(self) -> tuple:
        """
        """

        value = self.__callback(*self.__args)
        if self.__always_show or value:
            return (self.__show_label, value)

        return (False, None)

class InspectFormatter(object):
    """
    """

    @staticmethod
    def format(value: object) -> object:
        """
        """

        types = [
            Vec2,
            Vec3,
            Vec4,
            Point2, 
            Point3
        ]

        bool_values = {
            True: 'on',
            False: 'off'
        }

        if type(value) in types:
            s = '%0.2f/%0.2f' % (value.get_x(), value.get_y())
            if hasattr(value, 'get_z'):
                s += '/%0.2f' % value.get_z()
            if hasattr(value, 'get_w'):
                s += '/%0.2f' % value.get_w()
            value = s  
        elif type(value) in [bool]:
            value = bool_values[value]
        else:
            value = str(value)

        return value

 class Inspector(Runnable):
    """
    """

    def __init__(self, font_path: str= '', update_time: float = 0.1, formatter: object = InspectFormatter, show_labels: bool = False, parent: object = None, pos: tuple = (-1.32, 0.95), scale: float = 0.04, color: object = Vec4(1.0, 1.0, 1.0, 0.9)):
        super().__init__()
        self.formatter = formatter
        self.update_time = update_time
        self.show_labels = show_labels
        self.__last_update = 0.0
        self.objects = []

        font = None
        if path_exists(font_path):
            font = Font(font_path).get_text_font()

        self.onscreen_text = OnscreenText(
            fg=color, 
            align=TextNode.ALeft, 
            mayChange=True, 
            scale=scale, 
            font=font, 
            parent=parent, 
            pos=pos)

    def destroy(self) -> None:
        """
        Destroys the inspector object
        """

        Runnable.deactivate(self)
        self.onscreen_text.remove_node()
        self.objects = []

    def add_object(self, label: str, callback: object, always_show: bool = True, show_label: bool = False, *args) ->:
        """
        Adds a new object to the inspector
        """

        self.objects.append((label, InspectObject(always_show, show_label, callback, *args)))

    async def tick(self, dt: float) -> None:
        """
        Called once per tick to perform inspections
        """

        self.__last_update += dt
        if self.__last_update <= self.update_time:
            return

        txt = ''
        for label, obj in self.objects:
            show_label, value = obj.value
            if value is None:
                continue

            if show_label:
                txt += '%s: %s\n' % (label, self.formatter.format(value))
            else:
                txt += '%s\n' % self.formatter.format(value)

        self.onscreen_text.set_text(txt)
        self.__last_update = 0.0


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from panda3d_gemstone.framework.configurable import Configurable
from panda3d_gemstone.framework.runnable import Runnable
from panda3d_gemstone.framework.controller import ControllerChainMixIn
from panda3d_gemstone.controllers.avatar import AvatarController
from panda3d_gemstone.engine.model import AnimatedModel


class Avatar(AnimatedModel, Configurable, Runnable, ControllerChainMixIn):
    """
    Represents an avatar in the Gemstone framework
    """

    def __init__(self, config_path: str):
        Configurable.__init__(self, config_path)
        self.controller = None
        name = self.pop('name')
        self.__setup_model_path_and_controller()
        AnimatedModel.__init__(self, self.model_path, have_anim_in_node=False)
        if name:
            self.set_name(name)

        Runnable.__init__(self, priority=5)

    def destroy(self) -> None:
        """
        Called on avatar destruction
        """

        if self.controller:
            self.controller.destroy()
        self.controller = None

        self.deactive()
        AnimatedModel.destroy(self)

    def make_copy_to(self, destination: object, copy_controller_chain: bool = True, callback: object = None, extra_args: list = [], task_chain_name: str = None) -> None:
        """
        """

        super(AnimatedModel, self).make_copy_to(destination, callback=callback, extra_args=extra_args, task_chain_name=task_chain_name)
        destination.set_name(self.get_name())
        destination.model_path = self.model_path

        if copy_controller_chain:
            destination.__setup_model_path_and_controller()

    def configure_controller(self, config_path: str = None) -> None:
        """
        """

        if config_path is None:
            config_path = self.path

        configuration = Configurable(config_path)
        controller = self.setup_controller_chain(configuration.pop('controller'), config_path)

        if isinstance(controller, AvatarController) or controller is None:
            self.controller = controller
        else:
            self.notify.warning('Failed to configure controller: %s does not inherit AvatarController!' % controller.__class__.__name__)
            self.controller = None

    def __setup_model_path_and_controller(self) -> None:
        """
        """

        self.model_path = self.pop('model')
        self.configure_controller()
        self.pop('controller')
        self.initialize()

    def get_controller(self) -> object:
        """
        """

        return self.controller

    def activate(self) -> None:
        """
        Activates the avatar object
        """

        if self.is_activated():
            Runnable.deactivate(self)

        if self.controller:
            self.controller.foreach_call('activate')
        
        return Runnable.activate(self)

    def deactivate(self) -> None:
        """
        Deactivates the avatar object
        """

        if self.is_activated():
            return super(Runnable, self).dectivate()

        if self.controller:
            self.controller.foreach_call('dectivate')
        
        return super(Runnable, self).dectivate()

    def reload(self) -> None:
        """
        Reloads the avatar object
        """

        super(Configurable, self).reload()
        self.__setup_model_path_and_controller()
        super(AnimatedModel, self).reload(True)

    def relooad_model(self, model_path: str) -> None:
        """
        Reloads the avatar's model
        """

        self.model_path = model_path
        super(AnimatedModel, self).__init__(self.model_path)

    def get_state(self, name: str) -> object:
        """
        Retrieves an avatar's state
        """

        if self.controller:
            return self.controller.get_state(name)

        return False

    def get_running(self) -> bool:
        """
        Retrieves the avatar's running state
        """

        if self.controller:
            return self.controller.get_running()

        return False

    def set_running(self, running: bool) -> None:
        """
        Sets the avatar's running state
        """

        if self.controller:
            self.controller.set_running(running)

    async def update_controller(self, nodepath: object, dt: float) -> None:
        """
        Called once per frame to perform controller update operations
        """

        if not self.controller or not nodepath:
            return
        
        np = await self.controller.update(nodepath, dt)
        nodepath.set_quat(np.get_quat())
        nodepath.set_fluid_pos(np.get_pos())

    async def tick(self, dt: float) -> None:
        """
        Called once per frame to perform tick operations
        """

        if not self.controller:
            return

        await self.update_controller(self, dt)

    def play_emote(self, anim: object) -> None:
        """
        Plays an emotion animation on the avatar
        """

        if not self.controller:
            return

        self.controller.play_emote(anim)


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from PyQt5.QtGui import *

from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.framework.registry import ClassRegistry

class QObjectTreeView(QTreeView):
    """
    """

    def __init__(self, *args, **kwargs):
        super(QTreeView, self).__init__(*args, **kwargs) 

    def build_tree(self, asset_tree: dict) -> None:
        """
        """

        root_model = QStandardItemModel()
        self.setModel(root_model)
        self._populate_tree(asset_tree, root_model.invisibleRootItem())
        root_model.setHeaderData(0, Qt.Horizontal, 'Available Objects')

        self.setDragDropMode(QAbstractItemView.DragOnly)
        self.setDragEnabled(True)
        self.setAcceptDrops(False)

    def _populate_tree(self, children: object, parent: object) -> None:
        """
        """

        for child in sorted(children):
            child_item = QStandardItem(child)
            parent.appendRow(child_item)
            if isinstance(children, dict):
                self._populate_tree(children[child], child_item)

class QObjectLibraryWidget(QDockWidget, InternalObject):
    """
    Object library tree for the current Gemstone application. Auto populates
    from the application's class registry
    """

    def __init__(self, *args, **kwargs):
        super(QDockWidget, self).__init__('Object Library', *args, **kwargs)
        super(InternalObject, self).__init__()
        self._registry = ClassRegistry.get_singleton()
        self.build_widget()

    def build_widget(self):
        """
        Builds the layer properties widget
        """

        # Build widget object tree
        object_tree = {}
        for class_name, class_data in self._registry:
            _, import_path, _, meta = class_data
            path_parts = import_path.split('.')
            sub_name = path_parts[1].capitalize()
            root_name = path_parts[0].capitalize()

            editor_object = meta.get('editor_object', False)    
            #if not editor_object:
            #    self.notify.warning(meta)
            #    continue

            if root_name not in object_tree:
                object_tree[root_name] = {}

            if sub_name not in object_tree[root_name]:
                object_tree[root_name][sub_name] = []

            object_tree[root_name][sub_name].append(class_name)

        self.tree = QObjectTreeView(self)
        self.tree.build_tree(object_tree)
        self.setWidget(self.tree)


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from panda3d_gemstone.framework.internal_object import InternalObject

class PlayerGroup(InternalObject): 
    """
    Represents a player group in the Gemstone application
    """

    def __init__(self):
        InternalObject.__init__(self)

        self.__leader = None
        self.__regular_members = []

    @property
    def group_leader(self) -> object:
        """
        Current group leader
        """

        return self.__leader

    @property
    def group_members(self) -> list:
        """
        All members of the group
        """

        return self.get_members()

    @property
    def regular_members(self) -> list:
        """
        Regular members of the group
        """

        return self.__regular_members

    def empty(self) -> None:
        """
        Empties the group object
        """

        self.__leader = None
        self.__regular_members = []

    def destroy(self) -> None:
        """
        Destroys the group object
        """

        InternalObject.destroy(self)
        self.empty()

    def is_leader(self, member: object) -> bool:
        """
        Returns true if the member is the group leader
        """

        return member == self.__leader

    def is_member(self, member: object) -> bool:
        """
        Returns true if the member is in the group
        """

        return self.is_leader(member) or self.is_regular_member(member)

    def is_regular_member(self, member: object) -> bool:
        """
        Returns true if the member is a regular member of the group
        """

        return member in self.__regular_members

    def get_members(self) -> list:
        """
        Returns all members in the team
        """

        group = []
        if self.__leader:
            groups.append(self.__leader)

        group.extend(self.__regular_members)
        return group

    def get_leader(self) -> object:
        """
        Returns the group's leader
        """

        return self.__leader

    def has_leader(self) -> bool:
        """
        Returns true if the group has a leader
        """

        return self.get_leader() != None

    def remove(self, member: object) -> None:
        """
        Removes the member from the group
        """

        if member in self.__regular_members:
            self.__regular_members.remove(member)
        elif member == self.__leader:
            self.__leader = None
        else:
            self.notify.warning('Failed to remove member from group. %s is not a member of the group' % str(member))

    def add(self, member: object, leader: bool = False) -> None:
        """
        Adds the member to the group. Optionally as the party leader
        """

        if self.is_member(member):
            self.notify.warning('Failed to add member to group. %s is already a member of the group' % str(member))
            return

        if leader:
            if self.has_leader():
                self.notify.warning('Failed to add member to group as leader. Group already has a leader.')
            else:
                self.__leader = member
        else:
            self.__regular_members.append(member)
        


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import importlib.util
import builtins
import sys
import os

from direct.showbase.MessengerGlobal import messenger

dev = False

def __get_base_executable_name() -> str:
    """
    Returns the base executable name
    """

    basename = os.path.basename(sys.argv[0])
    if basename == '-m':
        basename = os.environ.get('APP_NAME', 'Gemstone-App')

    basename = os.path.splitext(basename)[0]
    return basename

executable_name = __get_base_executable_name()

def is_venv() -> bool:
    """
    Returns true if the application is being run inside
    a virtual environment
    """

    real_prefix = hasattr(sys, 'real_prefix')
    base_prefix = hasattr(sys, 'base_prefix') and sys.base_prefix != sys.prefix

    return real_prefix or base_prefix

def is_frozen() -> bool:
    """
    Returns true if the application is being run from within
    a frozen Python environment
    """

    spec = importlib.util.find_spec(__name__)
    return spec is not None and spec.origin is not None

def is_interactive() -> bool:
    """
    Returns true if the application is being run from an
    interactive command prompt
    """

    import sys
    return hasattr(sys, 'ps1') and hasattr(sys, 'ps2')

def is_developer_build() -> bool:
    """
    Returns true if the application is currently
    running as a developer build
    """
    
    return (dev or is_interactive()) and not is_frozen()

def has_thirdparty(import_string) -> bool:
    """
    Returns true if the requested import is found in 
    the environment
    """

    import importlib
    found = False
    try:
        module = importlib.import_module(import_string)
        found = True
    except:
        pass

    return found

def has_render_pipeline_support() -> bool:
    """
    Returns true if the application has tobspr's 
    Render Pipeline module available in the environment
    """

    return has_thirdparty('rpcore')

def __get_module() -> object:
    """
    Returns the runtime module's object instance
    """

    return sys.modules[__name__]

def __has_variable(variable_name: str) -> bool:
    """
    Returns true if the runtime module has the requested variable name defined.
    Is served out via the custom __getattr__ function as has_x() method names
    """

    module = __get_module()
    defined = hasattr(module, variable_name)
    found = False

    if defined:
        attr = getattr(module, variable_name)
        found = attr != None

    return found

def __get_variable(variable_name: str) -> object:
    """
    Returns the requested variable from the runtime module if it exists.
    Otherwise returning NoneType
    """

    if not __has_variable(variable_name):
        return None

    module = __get_module()
    return getattr(module, variable_name)

def __getattr__(key: str) -> object:
    """
    Custom get attribute handler for allowing access to the has_x method names
    of the Gemstone runtime module. Also exposes the builtins module
    for the legacy Panda3d builtins provided by the ShowBase instance
    """

    result = None
    is_has_method = key.startswith('has_')
    is_get_method = key.startswith('get_')

    if len(key) > 4:
        variable_name = key[4:]
    else:
        variable_name = key

    if is_has_method:
        return lambda: __has_variable(variable_name)
    elif is_get_method:
        return lambda: __get_variable(variable_name)
    elif hasattr(builtins, key):
        return getattr(builtins, key)

    if not result:
        raise AttributeError('runtime module has no attribute: %s' % key)

    return result


